@@ -2,69 +2,69 @@ from __future__ import annotations
 
 import contextlib
 import operator
-from typing import TYPE_CHECKING
-from typing import Any
-from typing import Callable
-from typing import Literal
-from typing import Sequence
-from typing import cast
+from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence, cast
 
-from duckdb import CoalesceOperator
-from duckdb import FunctionExpression
-from duckdb import StarExpression
+from duckdb import CoalesceOperator, FunctionExpression, StarExpression
 from duckdb.typing import DuckDBPyType
 
 from narwhals._compliant import LazyExpr
+from narwhals._compliant.window import UnorderableWindowInputs, WindowInputs
 from narwhals._duckdb.expr_dt import DuckDBExprDateTimeNamespace
 from narwhals._duckdb.expr_list import DuckDBExprListNamespace
 from narwhals._duckdb.expr_str import DuckDBExprStringNamespace
 from narwhals._duckdb.expr_struct import DuckDBExprStructNamespace
-from narwhals._duckdb.utils import UnorderableWindowInputs
-from narwhals._duckdb.utils import WindowInputs
-from narwhals._duckdb.utils import col
-from narwhals._duckdb.utils import ensure_type
-from narwhals._duckdb.utils import generate_order_by_sql
-from narwhals._duckdb.utils import generate_partition_by_sql
-from narwhals._duckdb.utils import lit
-from narwhals._duckdb.utils import narwhals_to_native_dtype
-from narwhals._duckdb.utils import when
+from narwhals._duckdb.utils import (
+    col,
+    generate_order_by_sql,
+    generate_partition_by_sql,
+    lit,
+    narwhals_to_native_dtype,
+    when,
+)
 from narwhals._expression_parsing import ExprKind
-from narwhals.utils import Implementation
-from narwhals.utils import not_implemented
-from narwhals.utils import requires
+from narwhals.utils import Implementation, not_implemented, requires
 
 if TYPE_CHECKING:
-    import duckdb
+    from duckdb import Expression
     from typing_extensions import Self
 
-    from narwhals._compliant.typing import AliasNames
-    from narwhals._compliant.typing import EvalNames
-    from narwhals._compliant.typing import EvalSeries
+    from narwhals._compliant.typing import (
+        AliasNames,
+        EvalNames,
+        EvalSeries,
+        UnorderableWindowFunction,
+        WindowFunction,
+    )
     from narwhals._duckdb.dataframe import DuckDBLazyFrame
     from narwhals._duckdb.namespace import DuckDBNamespace
-    from narwhals._duckdb.typing import UnorderableWindowFunction
-    from narwhals._duckdb.typing import WindowFunction
     from narwhals._expression_parsing import ExprMetadata
     from narwhals.dtypes import DType
-    from narwhals.typing import FillNullStrategy
-    from narwhals.typing import NonNestedLiteral
-    from narwhals.typing import NumericLiteral
-    from narwhals.typing import RankMethod
-    from narwhals.typing import RollingInterpolationMethod
-    from narwhals.typing import TemporalLiteral
-    from narwhals.utils import Version
-    from narwhals.utils import _FullContext
+    from narwhals.typing import (
+        FillNullStrategy,
+        NonNestedLiteral,
+        NumericLiteral,
+        RankMethod,
+        RollingInterpolationMethod,
+        TemporalLiteral,
+    )
+    from narwhals.utils import Version, _FullContext
+
+    DuckDBWindowInputs = WindowInputs[Expression]
+    DuckDBUnorderableWindowInputs = UnorderableWindowInputs[Expression]
+    DuckDBWindowFunction = WindowFunction[Expression]
+    DuckDBUnorderableWindowFunction = UnorderableWindowFunction[Expression]
+
 
 with contextlib.suppress(ImportError):  # requires duckdb>=1.3.0
-    from duckdb import SQLExpression  # type: ignore[attr-defined, unused-ignore]
+    from duckdb import SQLExpression
 
 
-class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
+class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "Expression"]):
     _implementation = Implementation.DUCKDB
 
     def __init__(
         self,
-        call: EvalSeries[DuckDBLazyFrame, duckdb.Expression],
+        call: EvalSeries[DuckDBLazyFrame, Expression],
         *,
         evaluate_output_names: EvalNames[DuckDBLazyFrame],
         alias_output_names: AliasNames | None,
@@ -79,13 +79,13 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         self._metadata: ExprMetadata | None = None
 
         # This can only be set by `_with_window_function`.
-        self._window_function: WindowFunction | None = None
+        self._window_function: DuckDBWindowFunction | None = None
 
         # These can only be set by `_with_unorderable_window_function`
-        self._unorderable_window_function: UnorderableWindowFunction | None = None
-        self._previous_call: EvalSeries[DuckDBLazyFrame, duckdb.Expression] | None = None
+        self._unorderable_window_function: DuckDBUnorderableWindowFunction | None = None
+        self._previous_call: EvalSeries[DuckDBLazyFrame, Expression] | None = None
 
-    def __call__(self, df: DuckDBLazyFrame) -> Sequence[duckdb.Expression]:
+    def __call__(self, df: DuckDBLazyFrame) -> Sequence[Expression]:
         return self._call(df)
 
     def __narwhals_expr__(self) -> None: ...
@@ -103,14 +103,12 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         *,
         reverse: bool,
         func_name: Literal["sum", "max", "min", "count", "product"],
-    ) -> WindowFunction:
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(
-                *window_inputs.order_by, ascending=not reverse
-            )
-            partition_by_sql = generate_partition_by_sql(*window_inputs.partition_by)
+    ) -> DuckDBWindowFunction:
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=not reverse)
+            partition_by_sql = generate_partition_by_sql(*inputs.partition_by)
             sql = (
-                f"{func_name} ({window_inputs.expr}) over ({partition_by_sql} {order_by_sql} "
+                f"{func_name} ({inputs.expr}) over ({partition_by_sql} {order_by_sql} "
                 "rows between unbounded preceding and current row)"
             )
             return SQLExpression(sql)  # type: ignore[no-any-return, unused-ignore]
@@ -125,9 +123,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         window_size: int,
         min_samples: int,
         ddof: int | None = None,
-    ) -> WindowFunction:
-        ensure_type(window_size, int, type(None))
-        ensure_type(min_samples, int)
+    ) -> DuckDBWindowFunction:
         supported_funcs = ["sum", "mean", "std", "var"]
         if center:
             half = (window_size - 1) // 2
@@ -138,9 +134,9 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
             start = f"{window_size - 1} preceding"
             end = "current row"
 
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(*window_inputs.order_by, ascending=True)
-            partition_by_sql = generate_partition_by_sql(*window_inputs.partition_by)
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=True)
+            partition_by_sql = generate_partition_by_sql(*inputs.partition_by)
             window = f"({partition_by_sql} {order_by_sql} rows between {start} and {end})"
             if func_name in {"sum", "mean"}:
                 func_: str = func_name
@@ -158,9 +154,9 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
             else:  # pragma: no cover
                 msg = f"Only the following functions are supported: {supported_funcs}.\nGot: {func_name}."
                 raise ValueError(msg)
-            condition_sql = f"count({window_inputs.expr}) over {window} >= {min_samples}"
+            condition_sql = f"count({inputs.expr}) over {window} >= {min_samples}"
             condition = SQLExpression(condition_sql)
-            value = SQLExpression(f"{func_}({window_inputs.expr}) over {window}")
+            value = SQLExpression(f"{func_}({inputs.expr}) over {window}")
             return when(condition, value)
 
         return func
@@ -171,19 +167,8 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         if self._backend_version < (1, 3):
             msg = "At least version 1.3 of DuckDB is required for binary operations between aggregates and columns."
             raise NotImplementedError(msg)
-
         template = "{expr} over ()"
-
-        def func(df: DuckDBLazyFrame) -> Sequence[duckdb.Expression]:
-            return [SQLExpression(template.format(expr=expr)) for expr in self(df)]
-
-        return self.__class__(
-            func,
-            evaluate_output_names=self._evaluate_output_names,
-            alias_output_names=self._alias_output_names,
-            backend_version=self._backend_version,
-            version=self._version,
-        )
+        return self._with_callable(lambda expr: SQLExpression(template.format(expr=expr)))
 
     @classmethod
     def from_column_names(
@@ -193,7 +178,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         *,
         context: _FullContext,
     ) -> Self:
-        def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+        def func(df: DuckDBLazyFrame) -> list[Expression]:
             return [col(name) for name in evaluate_column_names(df)]
 
         return cls(
@@ -206,7 +191,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
 
     @classmethod
     def from_column_indices(cls, *column_indices: int, context: _FullContext) -> Self:
-        def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+        def func(df: DuckDBLazyFrame) -> list[Expression]:
             columns = df.columns
             return [col(columns[i]) for i in column_indices]
 
@@ -219,7 +204,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         )
 
     def _with_callable(
-        self, call: Callable[..., duckdb.Expression], /, **expressifiable_args: Self | Any
+        self, call: Callable[..., Expression], /, **expressifiable_args: Self | Any
     ) -> Self:
         """Create expression from callable.
 
@@ -230,7 +215,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
                 as expressions (e.g. in `nw.col('a').is_between('b', 'c')`)
         """
 
-        def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+        def func(df: DuckDBLazyFrame) -> list[Expression]:
             native_series_list = self(df)
             other_native_series = {
                 key: df._evaluate_expr(value) if self._is_expr(value) else lit(value)
@@ -258,7 +243,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
             version=self._version,
         )
 
-    def _with_window_function(self, window_function: WindowFunction) -> Self:
+    def _with_window_function(self, window_function: DuckDBWindowFunction) -> Self:
         result = self.__class__(
             self._call,
             evaluate_output_names=self._evaluate_output_names,
@@ -271,8 +256,8 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
 
     def _with_unorderable_window_function(
         self,
-        unorderable_window_function: UnorderableWindowFunction,
-        previous_call: EvalSeries[DuckDBLazyFrame, duckdb.Expression],
+        unorderable_window_function: DuckDBUnorderableWindowFunction,
+        previous_call: EvalSeries[DuckDBLazyFrame, Expression],
     ) -> Self:
         result = self.__class__(
             self._call,
@@ -286,99 +271,97 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         return result
 
     @classmethod
-    def _alias_native(cls, expr: duckdb.Expression, name: str) -> duckdb.Expression:
+    def _alias_native(cls, expr: Expression, name: str) -> Expression:
         return expr.alias(name)
 
     def __and__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input & other, other=other)
+        return self._with_callable(lambda expr, other: expr & other, other=other)
 
     def __or__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input | other, other=other)
+        return self._with_callable(lambda expr, other: expr | other, other=other)
 
     def __add__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input + other, other=other)
+        return self._with_callable(lambda expr, other: expr + other, other=other)
 
     def __truediv__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input / other, other=other)
+        return self._with_callable(lambda expr, other: expr / other, other=other)
 
     def __rtruediv__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: other.__truediv__(_input), other=other
+            lambda expr, other: other.__truediv__(expr), other=other
         ).alias("literal")
 
     def __floordiv__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: _input.__floordiv__(other), other=other
+            lambda expr, other: expr.__floordiv__(other), other=other
         )
 
     def __rfloordiv__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: other.__floordiv__(_input), other=other
+            lambda expr, other: other.__floordiv__(expr), other=other
         ).alias("literal")
 
     def __mod__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(
-            lambda _input, other: _input.__mod__(other), other=other
-        )
+        return self._with_callable(lambda expr, other: expr.__mod__(other), other=other)
 
     def __rmod__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: other.__mod__(_input), other=other
+            lambda expr, other: other.__mod__(expr), other=other
         ).alias("literal")
 
     def __sub__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input - other, other=other)
+        return self._with_callable(lambda expr, other: expr - other, other=other)
 
     def __rsub__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: other.__sub__(_input), other=other
+            lambda expr, other: other.__sub__(expr), other=other
         ).alias("literal")
 
     def __mul__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input * other, other=other)
+        return self._with_callable(lambda expr, other: expr * other, other=other)
 
     def __pow__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input**other, other=other)
+        return self._with_callable(lambda expr, other: expr**other, other=other)
 
     def __rpow__(self, other: DuckDBExpr) -> Self:
         return self._with_callable(
-            lambda _input, other: other.__pow__(_input), other=other
+            lambda expr, other: other.__pow__(expr), other=other
         ).alias("literal")
 
     def __lt__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input < other, other=other)
+        return self._with_callable(lambda expr, other: expr < other, other=other)
 
     def __gt__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input > other, other=other)
+        return self._with_callable(lambda expr, other: expr > other, other=other)
 
     def __le__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input <= other, other=other)
+        return self._with_callable(lambda expr, other: expr <= other, other=other)
 
     def __ge__(self, other: DuckDBExpr) -> Self:
-        return self._with_callable(lambda _input, other: _input >= other, other=other)
+        return self._with_callable(lambda expr, other: expr >= other, other=other)
 
     def __eq__(self, other: DuckDBExpr) -> Self:  # type: ignore[override]
-        return self._with_callable(lambda _input, other: _input == other, other=other)
+        return self._with_callable(lambda expr, other: expr == other, other=other)
 
     def __ne__(self, other: DuckDBExpr) -> Self:  # type: ignore[override]
-        return self._with_callable(lambda _input, other: _input != other, other=other)
+        return self._with_callable(lambda expr, other: expr != other, other=other)
 
     def __invert__(self) -> Self:
-        invert = cast("Callable[..., duckdb.Expression]", operator.invert)
+        invert = cast("Callable[..., Expression]", operator.invert)
         return self._with_callable(invert)
 
     def abs(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("abs", _input))
+        return self._with_callable(lambda expr: FunctionExpression("abs", expr))
 
     def mean(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("mean", _input))
+        return self._with_callable(lambda expr: FunctionExpression("mean", expr))
 
     def skew(self) -> Self:
-        def func(_input: duckdb.Expression) -> duckdb.Expression:
-            count = FunctionExpression("count", _input)
+        def func(expr: Expression) -> Expression:
+            count = FunctionExpression("count", expr)
             # Adjust population skewness by correction factor to get sample skewness
             sample_skewness = (
-                FunctionExpression("skewness", _input)
+                FunctionExpression("skewness", expr)
                 * (count - lit(2))
                 / FunctionExpression("sqrt", count * (count - lit(1)))
             )
@@ -391,20 +374,20 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         return self._with_callable(func)
 
     def median(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("median", _input))
+        return self._with_callable(lambda expr: FunctionExpression("median", expr))
 
     def all(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("bool_and", _input))
+        return self._with_callable(lambda expr: FunctionExpression("bool_and", expr))
 
     def any(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("bool_or", _input))
+        return self._with_callable(lambda expr: FunctionExpression("bool_or", expr))
 
     def quantile(
         self, quantile: float, interpolation: RollingInterpolationMethod
     ) -> Self:
-        def func(_input: duckdb.Expression) -> duckdb.Expression:
+        def func(expr: Expression) -> Expression:
             if interpolation == "linear":
-                return FunctionExpression("quantile_cont", _input, lit(quantile))
+                return FunctionExpression("quantile_cont", expr, lit(quantile))
             msg = "Only linear interpolation methods are supported for DuckDB quantile."
             raise NotImplementedError(msg)
 
@@ -415,17 +398,17 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         lower_bound: Self | NumericLiteral | TemporalLiteral | None,
         upper_bound: Self | NumericLiteral | TemporalLiteral | None,
     ) -> Self:
-        def _clip_lower(_input: duckdb.Expression, lower_bound: Any) -> duckdb.Expression:
-            return FunctionExpression("greatest", _input, lower_bound)
+        def _clip_lower(expr: Expression, lower_bound: Any) -> Expression:
+            return FunctionExpression("greatest", expr, lower_bound)
 
-        def _clip_upper(_input: duckdb.Expression, upper_bound: Any) -> duckdb.Expression:
-            return FunctionExpression("least", _input, upper_bound)
+        def _clip_upper(expr: Expression, upper_bound: Any) -> Expression:
+            return FunctionExpression("least", expr, upper_bound)
 
         def _clip_both(
-            _input: duckdb.Expression, lower_bound: Any, upper_bound: Any
-        ) -> duckdb.Expression:
+            expr: Expression, lower_bound: Any, upper_bound: Any
+        ) -> Expression:
             return FunctionExpression(
-                "greatest", FunctionExpression("least", _input, upper_bound), lower_bound
+                "greatest", FunctionExpression("least", expr, upper_bound), lower_bound
             )
 
         if lower_bound is None:
@@ -437,39 +420,39 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         )
 
     def sum(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("sum", _input))
+        return self._with_callable(lambda expr: FunctionExpression("sum", expr))
 
     def n_unique(self) -> Self:
-        def func(_input: duckdb.Expression) -> duckdb.Expression:
+        def func(expr: Expression) -> Expression:
             # https://stackoverflow.com/a/79338887/4451315
             return FunctionExpression(
-                "array_unique", FunctionExpression("array_agg", _input)
+                "array_unique", FunctionExpression("array_agg", expr)
             ) + FunctionExpression(
-                "max", when(_input.isnotnull(), lit(0)).otherwise(lit(1))
+                "max", when(expr.isnotnull(), lit(0)).otherwise(lit(1))
             )
 
         return self._with_callable(func)
 
     def count(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("count", _input))
+        return self._with_callable(lambda expr: FunctionExpression("count", expr))
 
     def len(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("count"))
+        return self._with_callable(lambda _expr: FunctionExpression("count"))
 
     def std(self, ddof: int) -> Self:
         if ddof == 0:
             return self._with_callable(
-                lambda _input: FunctionExpression("stddev_pop", _input)
+                lambda expr: FunctionExpression("stddev_pop", expr)
             )
         if ddof == 1:
             return self._with_callable(
-                lambda _input: FunctionExpression("stddev_samp", _input)
+                lambda expr: FunctionExpression("stddev_samp", expr)
             )
 
-        def _std(_input: duckdb.Expression) -> duckdb.Expression:
-            n_samples = FunctionExpression("count", _input)
+        def _std(expr: Expression) -> Expression:
+            n_samples = FunctionExpression("count", expr)
             return (
-                FunctionExpression("stddev_pop", _input)
+                FunctionExpression("stddev_pop", expr)
                 * FunctionExpression("sqrt", n_samples)
                 / (FunctionExpression("sqrt", (n_samples - lit(ddof))))
             )
@@ -478,33 +461,27 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
 
     def var(self, ddof: int) -> Self:
         if ddof == 0:
-            return self._with_callable(
-                lambda _input: FunctionExpression("var_pop", _input)
-            )
+            return self._with_callable(lambda expr: FunctionExpression("var_pop", expr))
         if ddof == 1:
-            return self._with_callable(
-                lambda _input: FunctionExpression("var_samp", _input)
-            )
+            return self._with_callable(lambda expr: FunctionExpression("var_samp", expr))
 
-        def _var(_input: duckdb.Expression) -> duckdb.Expression:
-            n_samples = FunctionExpression("count", _input)
+        def _var(expr: Expression) -> Expression:
+            n_samples = FunctionExpression("count", expr)
             return (
-                FunctionExpression("var_pop", _input)
-                * n_samples
-                / (n_samples - lit(ddof))
+                FunctionExpression("var_pop", expr) * n_samples / (n_samples - lit(ddof))
             )
 
         return self._with_callable(_var)
 
     def max(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("max", _input))
+        return self._with_callable(lambda expr: FunctionExpression("max", expr))
 
     def min(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("min", _input))
+        return self._with_callable(lambda expr: FunctionExpression("min", expr))
 
     def null_count(self) -> Self:
         return self._with_callable(
-            lambda _input: FunctionExpression("sum", _input.isnull().cast("int")),
+            lambda expr: FunctionExpression("sum", expr.isnull().cast("int"))
         )
 
     @requires.backend_version((1, 3))
@@ -512,7 +489,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         if (window_function := self._window_function) is not None:
             assert order_by is not None  # noqa: S101
 
-            def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+            def func(df: DuckDBLazyFrame) -> list[Expression]:
                 return [
                     window_function(WindowInputs(expr, partition_by, order_by))
                     for expr in self._call(df)
@@ -522,7 +499,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         ) is not None:
             assert order_by is None  # noqa: S101
 
-            def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+            def func(df: DuckDBLazyFrame) -> list[Expression]:
                 assert self._previous_call is not None  # noqa: S101
                 return [
                     unorderable_window_function(
@@ -534,7 +511,7 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
             partition_by_sql = generate_partition_by_sql(*partition_by)
             template = f"{{expr}} over ({partition_by_sql})"
 
-            def func(df: DuckDBLazyFrame) -> list[duckdb.Expression]:
+            def func(df: DuckDBLazyFrame) -> list[Expression]:
                 return [
                     SQLExpression(template.format(expr=expr)) for expr in self._call(df)
                 ]
@@ -548,49 +525,44 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
         )
 
     def is_null(self) -> Self:
-        return self._with_callable(lambda _input: _input.isnull())
+        return self._with_callable(lambda expr: expr.isnull())
 
     def is_nan(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("isnan", _input))
+        return self._with_callable(lambda expr: FunctionExpression("isnan", expr))
 
     def is_finite(self) -> Self:
-        return self._with_callable(lambda _input: FunctionExpression("isfinite", _input))
+        return self._with_callable(lambda expr: FunctionExpression("isfinite", expr))
 
     def is_in(self, other: Sequence[Any]) -> Self:
         return self._with_callable(
-            lambda _input: FunctionExpression("contains", lit(other), _input)
+            lambda expr: FunctionExpression("contains", lit(other), expr)
         )
 
     def round(self, decimals: int) -> Self:
         return self._with_callable(
-            lambda _input: FunctionExpression("round", _input, lit(decimals))
+            lambda expr: FunctionExpression("round", expr, lit(decimals))
         )
 
     @requires.backend_version((1, 3))
     def shift(self, n: int) -> Self:
-        ensure_type(n, int)
-
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(*window_inputs.order_by, ascending=True)
-            partition_by_sql = generate_partition_by_sql(*window_inputs.partition_by)
-            sql = (
-                f"lag({window_inputs.expr}, {n}) over ({partition_by_sql} {order_by_sql})"
-            )
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=True)
+            partition_by_sql = generate_partition_by_sql(*inputs.partition_by)
+            sql = f"lag({inputs.expr}, {n}) over ({partition_by_sql} {order_by_sql})"
             return SQLExpression(sql)  # type: ignore[no-any-return, unused-ignore]
 
         return self._with_window_function(func)
 
     @requires.backend_version((1, 3))
     def is_first_distinct(self) -> Self:
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(*window_inputs.order_by, ascending=True)
-            if window_inputs.partition_by:
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=True)
+            if inputs.partition_by:
                 partition_by_sql = (
-                    generate_partition_by_sql(*window_inputs.partition_by)
-                    + f", {window_inputs.expr}"
+                    generate_partition_by_sql(*inputs.partition_by) + f", {inputs.expr}"
                 )
             else:
-                partition_by_sql = f"partition by {window_inputs.expr}"
+                partition_by_sql = f"partition by {inputs.expr}"
             sql = f"{FunctionExpression('row_number')} over({partition_by_sql} {order_by_sql})"
             return SQLExpression(sql) == lit(1)  # type: ignore[no-any-return, unused-ignore]
 
@@ -598,15 +570,14 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
 
     @requires.backend_version((1, 3))
     def is_last_distinct(self) -> Self:
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(*window_inputs.order_by, ascending=False)
-            if window_inputs.partition_by:
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=False)
+            if inputs.partition_by:
                 partition_by_sql = (
-                    generate_partition_by_sql(*window_inputs.partition_by)
-                    + f", {window_inputs.expr}"
+                    generate_partition_by_sql(*inputs.partition_by) + f", {inputs.expr}"
                 )
             else:
-                partition_by_sql = f"partition by {window_inputs.expr}"
+                partition_by_sql = f"partition by {inputs.expr}"
             sql = f"{FunctionExpression('row_number')} over({partition_by_sql} {order_by_sql})"
             return SQLExpression(sql) == lit(1)  # type: ignore[no-any-return, unused-ignore]
 
@@ -614,11 +585,11 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
 
     @requires.backend_version((1, 3))
     def diff(self) -> Self:
-        def func(window_inputs: WindowInputs) -> duckdb.Expression:
-            order_by_sql = generate_order_by_sql(*window_inputs.order_by, ascending=True)
-            partition_by_sql = generate_partition_by_sql(*window_inputs.partition_by)
-            sql = f"lag({window_inputs.expr}) over ({partition_by_sql} {order_by_sql})"
-            return window_inputs.expr - SQLExpression(sql)  # type: ignore[no-any-return, unused-ignore]
+        def func(inputs: DuckDBWindowInputs) -> Expression:
+            order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=True)
+            partition_by_sql = generate_partition_by_sql(*inputs.partition_by)
+            sql = f"lag({inputs.expr}) over ({partition_by_sql} {order_by_sql})"
+            return inputs.expr - SQLExpression(sql)  # type: ignore[no-any-return, unused-ignore]
 
         return self._with_window_function(func)
 
@@ -713,11 +684,9 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
                 msg = f"`fill_null` with `strategy={strategy}` is only available in 'duckdb>=1.3.0'."
                 raise NotImplementedError(msg)
 
-            def _fill_with_strategy(window_inputs: WindowInputs) -> duckdb.Expression:
-                order_by_sql = generate_order_by_sql(
-                    *window_inputs.order_by, ascending=True
-                )
-                partition_by_sql = generate_partition_by_sql(*window_inputs.partition_by)
+            def _fill_with_strategy(inputs: DuckDBWindowInputs) -> Expression:
+                order_by_sql = generate_order_by_sql(*inputs.order_by, ascending=True)
+                partition_by_sql = generate_partition_by_sql(*inputs.partition_by)
 
                 fill_func = "last_value" if strategy == "forward" else "first_value"
                 _limit = "unbounded" if limit is None else limit
@@ -727,29 +696,29 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
                     else f"current row and {_limit} following"
                 )
                 sql = (
-                    f"{fill_func}({window_inputs.expr} ignore nulls) over "
+                    f"{fill_func}({inputs.expr} ignore nulls) over "
                     f"({partition_by_sql} {order_by_sql} rows between {rows_between})"
                 )
                 return SQLExpression(sql)  # type: ignore[no-any-return, unused-ignore]
 
             return self._with_window_function(_fill_with_strategy)
 
-        def _fill_constant(_input: duckdb.Expression, value: Any) -> duckdb.Expression:
-            return CoalesceOperator(_input, value)
+        def _fill_constant(expr: Expression, value: Any) -> Expression:
+            return CoalesceOperator(expr, value)
 
         return self._with_callable(_fill_constant, value=value)
 
     def cast(self, dtype: DType | type[DType]) -> Self:
-        def func(_input: duckdb.Expression) -> duckdb.Expression:
+        def func(expr: Expression) -> Expression:
             native_dtype = narwhals_to_native_dtype(dtype, self._version)
-            return _input.cast(DuckDBPyType(native_dtype))
+            return expr.cast(DuckDBPyType(native_dtype))
 
         return self._with_callable(func)
 
     @requires.backend_version((1, 3))
     def is_unique(self) -> Self:
-        def func(_input: duckdb.Expression) -> duckdb.Expression:
-            sql = f"count(*) over (partition by {_input})"
+        def func(expr: Expression) -> Expression:
+            sql = f"count(*) over (partition by {expr})"
             return SQLExpression(sql) == lit(1)  # type: ignore[no-any-return, unused-ignore]
 
         return self._with_callable(func)
@@ -764,60 +733,55 @@ class DuckDBExpr(LazyExpr["DuckDBLazyFrame", "duckdb.Expression"]):
             func = FunctionExpression("row_number")
 
         def _rank(
-            _input: duckdb.Expression,
+            expr: Expression,
             *,
             descending: bool,
-            partition_by: Sequence[str | duckdb.Expression] | None = None,
-        ) -> duckdb.Expression:
+            partition_by: Sequence[str | Expression] | None = None,
+        ) -> Expression:
             order_by_sql = (
-                f"order by {_input} desc nulls last"
+                f"order by {expr} desc nulls last"
                 if descending
-                else f"order by {_input} asc nulls last"
+                else f"order by {expr} asc nulls last"
             )
             count_expr = FunctionExpression("count", StarExpression())
             if partition_by is not None:
                 window = f"{generate_partition_by_sql(*partition_by)} {order_by_sql}"
-                count_window = f"{generate_partition_by_sql(*partition_by, _input)}"
+                count_window = f"{generate_partition_by_sql(*partition_by, expr)}"
             else:
                 window = order_by_sql
-                count_window = generate_partition_by_sql(_input)
+                count_window = generate_partition_by_sql(expr)
             if method == "max":
-                expr = (
+                rank_expr = (
                     SQLExpression(f"{func} OVER ({window})")
                     + SQLExpression(f"{count_expr} over ({count_window})")
                     - lit(1)
                 )
             elif method == "average":
-                expr = SQLExpression(f"{func} OVER ({window})") + (
+                rank_expr = SQLExpression(f"{func} OVER ({window})") + (
                     SQLExpression(f"{count_expr} over ({count_window})") - lit(1)
                 ) / lit(2.0)
             else:
-                expr = SQLExpression(f"{func} OVER ({window})")
-            return when(_input.isnotnull(), expr)
+                rank_expr = SQLExpression(f"{func} OVER ({window})")
+            return when(expr.isnotnull(), rank_expr)
 
-        def _unpartitioned_rank(_input: duckdb.Expression) -> duckdb.Expression:
-            return _rank(_input, descending=descending)
+        def _unpartitioned_rank(expr: Expression) -> Expression:
+            return _rank(expr, descending=descending)
 
-        def _partitioned_rank(
-            window_inputs: UnorderableWindowInputs,
-        ) -> duckdb.Expression:
+        def _partitioned_rank(inputs: DuckDBUnorderableWindowInputs) -> Expression:
             return _rank(
-                window_inputs.expr,
-                descending=descending,
-                partition_by=window_inputs.partition_by,
+                inputs.expr, descending=descending, partition_by=inputs.partition_by
             )
 
         return self._with_callable(_unpartitioned_rank)._with_unorderable_window_function(
-            _partitioned_rank,
-            self._call,
+            _partitioned_rank, self._call
         )
 
     def log(self, base: float) -> Self:
-        def _log(_input: duckdb.Expression) -> duckdb.Expression:
-            log = FunctionExpression("log", _input)
+        def _log(expr: Expression) -> Expression:
+            log = FunctionExpression("log", expr)
             return (
-                when(_input < lit(0), lit(float("nan")))
-                .when(_input == lit(0), lit(float("-inf")))
+                when(expr < lit(0), lit(float("nan")))
+                .when(expr == lit(0), lit(float("-inf")))
                 .otherwise(log / FunctionExpression("log", lit(base)))
             )
 
